@@ -4,13 +4,16 @@
 
 #include "QXmppBlockingManager.h"
 
+#include "QXmppConstants_p.h"
 #include "QXmppIqHandling.h"
 #include "QXmppUtils.h"
 #include "QXmppUtils_p.h"
 
-#include <QDomElement>
+#include "Iq.h"
+#include "StringLiterals.h"
+#include "XmlWriter.h"
 
-constexpr QStringView XMLNS_BLOCKING = u"urn:xmpp:blocking";
+#include <QDomElement>
 
 using namespace QXmpp;
 using namespace QXmpp::Private;
@@ -24,101 +27,39 @@ static void makeUnique(T &vec)
     vec.erase(std::unique(vec.begin(), vec.end()), vec.end());
 }
 
-// IQ parsing helpers
-static QVector<QString> parseItems(const QDomElement &el)
-{
-    QVector<QString> jids;
-    for (const auto &item : iterChildElements(el, u"item")) {
-        jids.append(item.attribute(QStringLiteral("jid")));
-    }
-    return jids;
-}
-
-static void serializeItems(QXmlStreamWriter *writer, const QVector<QString> &jids)
-{
-    for (const auto &jid : jids) {
-        writer->writeStartElement(QSL65("item"));
-        writer->writeAttribute(QSL65("jid"), jid);
-        writer->writeEndElement();
-    }
-}
-
 // IQs
-class BlocklistIq : public QXmppIq
-{
-public:
-    QVector<QString> jids;
-
-    void parseElementFromChild(const QDomElement &el) override
-    {
-        jids = parseItems(el.firstChildElement());
-    }
-    void toXmlElementFromChild(QXmlStreamWriter *writer) const override
-    {
-        writer->writeStartElement(QSL65("blocklist"));
-        writer->writeDefaultNamespace(XMLNS_BLOCKING.toString());
-        serializeItems(writer, jids);
-        writer->writeEndElement();
-    }
-    static bool checkIqType(const QString &tagName, const QString &xmlns)
-    {
-        return tagName == u"blocklist" && xmlns == XMLNS_BLOCKING;
-    }
+enum BlockingAction {
+    List,
+    Block,
+    Unblock,
 };
 
-class BlockIq : public QXmppIq
-{
-public:
-    QVector<QString> jids;
-
-    explicit BlockIq(QVector<QString> jids = {})
-        : QXmppIq(QXmppIq::Set), jids(std::move(jids))
-    {
-    }
-
-    void parseElementFromChild(const QDomElement &el) override
-    {
-        jids = parseItems(el.firstChildElement());
-    }
-    void toXmlElementFromChild(QXmlStreamWriter *writer) const override
-    {
-        writer->writeStartElement(QSL65("block"));
-        writer->writeDefaultNamespace(XMLNS_BLOCKING.toString());
-        serializeItems(writer, jids);
-        writer->writeEndElement();
-    }
-    static bool checkIqType(const QString &tagName, const QString &xmlns)
-    {
-        return tagName == u"block" && xmlns == XMLNS_BLOCKING;
-    }
+template<>
+struct Enums::Data<BlockingAction> {
+    static constexpr auto Values = makeValues<BlockingAction>({
+        { List, u"blocklist" },
+        { Block, u"block" },
+        { Unblock, u"unblock" },
+    });
 };
 
-class UnblockIq : public QXmppIq
-{
-public:
+template<BlockingAction Action>
+struct Blocking {
     QVector<QString> jids;
 
-    explicit UnblockIq(QVector<QString> jids = {})
-        : QXmppIq(QXmppIq::Set), jids(std::move(jids))
-    {
-    }
+    static constexpr std::tuple XmlTag = { Enums::toString(Action), ns_blocking };
 
-    void parseElementFromChild(const QDomElement &el) override
+    bool parse(const QDomElement &el)
     {
-        jids = parseItems(el.firstChildElement());
+        if (elementXmlTag(el) == XmlTag) {
+            jids = parseSingleAttributeElements<QVector<QString>>(el, u"item", ns_blocking, u"jid"_s);
+            return true;
+        }
+        return false;
     }
-
-    void toXmlElementFromChild(QXmlStreamWriter *writer) const override
+    void toXml(XmlWriter &w) const
     {
-        writer->writeStartElement(QSL65("unblock"));
-        writer->writeDefaultNamespace(XMLNS_BLOCKING.toString());
-        serializeItems(writer, jids);
-        writer->writeEndElement();
-    }
-
-    static bool checkIqType(const QString &tagName, const QString &xmlns)
-    {
-        return tagName == u"unblock" && xmlns == XMLNS_BLOCKING;
+        w.write(Element { XmlTag, SingleAttributeElements { u"item", u"jid", jids } });
     }
 };
 
@@ -284,11 +225,16 @@ QXmppTask<QXmppBlockingManager::BlocklistResult> QXmppBlockingManager::fetchBloc
 
     // send IQ request, if this is the first request
     if (d->openFetchBlocklistPromises.size() == 1) {
-        client()->sendIq(BlocklistIq()).then(this, [this](QXmppClient::IqResult &&result) {
+        auto iq = CompatIq {
+            GetIq<Blocking<List>> {
+                generateSequentialStanzaId(),
+            },
+        };
+        client()->sendIq(std::move(iq)).then(this, [this](QXmppClient::IqResult &&result) {
             // parse into QXmppBlocklist/Error variant
-            auto blocklistResult = parseIq<BlocklistIq>(std::move(result), [](BlocklistIq &&iq) -> BlocklistResult {
-                return QXmppBlocklist(std::move(iq.jids));
-            });
+            auto blocklistResult = map<BlocklistResult>(
+                [](Blocking<List> &&list) { return QXmppBlocklist(list.jids); },
+                parseIqResponse<Blocking<List>>(std::move(result)));
 
             // store blocklist on success
             if (!d->blocklist) {
@@ -326,7 +272,10 @@ QXmppTask<QXmppBlockingManager::BlocklistResult> QXmppBlockingManager::fetchBloc
 ///
 QXmppTask<QXmppBlockingManager::Result> QXmppBlockingManager::block(QVector<QString> jids)
 {
-    return client()->sendGenericIq(BlockIq(std::move(jids)));
+    return client()->sendGenericIq(CompatIq {
+        SetIq<Blocking<Block>> {
+            generateSequentialStanzaId(), {}, {}, {}, { std::move(jids) } },
+    });
 }
 
 ///
@@ -344,13 +293,16 @@ QXmppTask<QXmppBlockingManager::Result> QXmppBlockingManager::block(QVector<QStr
 ///
 QXmppTask<QXmppBlockingManager::Result> QXmppBlockingManager::unblock(QVector<QString> jids)
 {
-    return client()->sendGenericIq(UnblockIq(std::move(jids)));
+    return client()->sendGenericIq(CompatIq {
+        SetIq<Blocking<Unblock>> {
+            generateSequentialStanzaId(), {}, {}, {}, { std::move(jids) } },
+    });
 }
 
 /// \cond
 QStringList QXmppBlockingManager::discoveryFeatures() const
 {
-    return { XMLNS_BLOCKING.toString() };
+    return { ns_blocking.toString() };
 }
 
 void QXmppBlockingManager::onRegistered(QXmppClient *client)
@@ -365,16 +317,7 @@ void QXmppBlockingManager::onUnregistered(QXmppClient *oldClient)
 
 bool QXmppBlockingManager::handleStanza(const QDomElement &stanza, const std::optional<QXmppE2eeMetadata> &)
 {
-    auto checkIqValidity = [this](QXmppIq::Type type, QStringView from) -> std::optional<StanzaError> {
-        // check type
-        if (type != QXmppIq::Set) {
-            return StanzaError {
-                StanzaError::Cancel,
-                StanzaError::FeatureNotImplemented,
-                QStringLiteral("Only IQs of type 'set' supported.")
-            };
-        }
-
+    auto checkIqValidity = [this](QStringView from) -> std::optional<StanzaError> {
         // check permissions
         // only server (user's account) is allowed to do blocklist pushes
         if (!from.isEmpty() && from != client()->configuration().jidBare()) {
@@ -382,7 +325,7 @@ bool QXmppBlockingManager::handleStanza(const QDomElement &stanza, const std::op
             return StanzaError {
                 StanzaError::Cancel,
                 StanzaError::Forbidden,
-                QStringLiteral("Forbidden.")
+                u"Forbidden."_s
             };
         }
 
@@ -390,7 +333,7 @@ bool QXmppBlockingManager::handleStanza(const QDomElement &stanza, const std::op
             return StanzaError {
                 StanzaError::Wait,
                 StanzaError::UnexpectedRequest,
-                QStringLiteral("Client is not subscribed to blocklist.")
+                u"Client is not subscribed to blocklist."_s
             };
         }
 
@@ -398,35 +341,35 @@ bool QXmppBlockingManager::handleStanza(const QDomElement &stanza, const std::op
         return {};
     };
 
-    auto handleBlock = [this, checkIqValidity](BlockIq &&iq) -> std::variant<QXmppIq, StanzaError> {
-        if (auto err = checkIqValidity(iq.type(), iq.from())) {
+    auto handleBlock = [this, checkIqValidity](SetIq<Blocking<Block>> &&iq) -> std::variant<QXmppIq, StanzaError> {
+        if (auto err = checkIqValidity(iq.from)) {
             return std::move(*err);
         }
 
         // store new jids
-        d->blocklist->append(iq.jids);
+        d->blocklist->append(iq.payload.jids);
         // assure blocklist has no duplicates
         makeUnique(*d->blocklist);
 
-        Q_EMIT blocked(iq.jids);
+        Q_EMIT blocked(iq.payload.jids);
         return QXmppIq(QXmppIq::Result);
     };
-    auto handleUnblock = [this, checkIqValidity](UnblockIq &&iq) -> std::variant<QXmppIq, StanzaError> {
-        if (auto err = checkIqValidity(iq.type(), iq.from())) {
+    auto handleUnblock = [this, checkIqValidity](SetIq<Blocking<Unblock>> &&iq) -> std::variant<QXmppIq, StanzaError> {
+        if (auto err = checkIqValidity(iq.from)) {
             return std::move(*err);
         }
 
         // remove jids
-        for (const auto &jid : iq.jids) {
+        for (const auto &jid : std::as_const(iq.payload.jids)) {
             d->blocklist->removeOne(jid);
         }
 
-        Q_EMIT unblocked(iq.jids);
+        Q_EMIT unblocked(iq.payload.jids);
         return QXmppIq(QXmppIq::Result);
     };
 
     // e2ee is not supported (not needed with local server)
-    return handleIqRequests<BlockIq, UnblockIq>(stanza, client(), overloaded { handleBlock, handleUnblock });
+    return handleIqRequests<SetIq<Blocking<Block>>, SetIq<Blocking<Unblock>>>(stanza, client(), overloaded { handleBlock, handleUnblock });
 }
 /// \endcond
 
