@@ -1,10 +1,12 @@
 // SPDX-FileCopyrightText: 2009 Manjeet Dahiya <manjeetdahiya@gmail.com>
 // SPDX-FileCopyrightText: 2019 Linus Jahn <lnj@kaidan.im>
+// SPDX-FileCopyrightText: 2023 Melvin Keskin <melvo@olomono.de>
 //
 // SPDX-License-Identifier: LGPL-2.1-or-later
 
 #include "QXmppClient.h"
 
+#include "QXmppAsync_p.h"
 #include "QXmppClientExtension.h"
 #include "QXmppClient_p.h"
 #include "QXmppConstants_p.h"
@@ -13,7 +15,6 @@
 #include "QXmppE2eeExtension.h"
 #include "QXmppE2eeMetadata.h"
 #include "QXmppEntityTimeManager.h"
-#include "QXmppFutureUtils_p.h"
 #include "QXmppLogger.h"
 #include "QXmppMessage.h"
 #include "QXmppMessageHandler.h"
@@ -26,6 +27,7 @@
 #include "QXmppVCardManager.h"
 #include "QXmppVersionManager.h"
 
+#include "Algorithms.h"
 #include "StringLiterals.h"
 #include "XmppSocket.h"
 
@@ -123,6 +125,13 @@ QStringList QXmppClientPrivate::discoveryFeatures()
 
 void QXmppClientPrivate::onErrorOccurred(const QString &text, const QXmppOutgoingClient::ConnectionError &err, QXmppClient::Error oldError)
 {
+    // Skip stream errors that are valid during special procedures such as account
+    // creation/deletion.
+    if (const auto streamError = std::get_if<QXmpp::StreamError>(&err);
+        streamError && ignoredStreamErrors.contains(*streamError)) {
+        return;
+    }
+
     if (q->configuration().autoReconnectionEnabled()) {
         if (oldError == QXmppClient::XmppStreamError) {
             // if we receive a resource conflict, inhibit reconnection
@@ -140,10 +149,7 @@ void QXmppClientPrivate::onErrorOccurred(const QString &text, const QXmppOutgoin
 
     // notify managers
     Q_EMIT q->error(oldError);
-    Q_EMIT q->errorOccurred(QXmppError {
-        text,
-        visit([](const auto &value) { return std::any(value); }, err),
-    });
+    Q_EMIT q->errorOccurred(QXmppError { text, into<std::any>(std::move(err)) });
 }
 /// \endcond
 
@@ -308,9 +314,6 @@ QXmppClient::QXmppClient(InitialExtensions initialExtensions, QObject *parent)
     connect(d->stream, &QXmppOutgoingClient::sslErrors,
             this, &QXmppClient::sslErrors);
 
-    connect(d->stream->socket(), &QAbstractSocket::stateChanged,
-            this, &QXmppClient::_q_socketStateChanged);
-
     connect(d->stream, &QXmppOutgoingClient::connected,
             this, &QXmppClient::_q_streamConnected);
 
@@ -321,6 +324,8 @@ QXmppClient::QXmppClient(InitialExtensions initialExtensions, QObject *parent)
         d->onErrorOccurred(text, error, oldError);
     });
 
+    connect(&d->stream->xmppSocket(), &XmppSocket::internalSocketStateChanged, this, &QXmppClient::onInternalSocketStateChanged);
+
     // reconnection
     d->reconnectionTimer = new QTimer(this);
     d->reconnectionTimer->setSingleShot(true);
@@ -635,7 +640,7 @@ QXmppTask<QXmppClient::IqResult> QXmppClient::sendSensitiveIq(QXmppIq &&iq, cons
             std::visit(overloaded {
                            [&](std::unique_ptr<QXmppIq> &&iq) {
                                // success (encrypted)
-                               d->stream->sendIq(std::move(*iq)).then(this, [this, p = std::move(p)](auto &&result) mutable {
+                               d->stream->sendIq(std::move(*iq)).then(this, [this, p = std::move(p)](IqResult &&result) mutable {
                                    // iq sent, response received
                                    std::visit(overloaded {
                                                   [&](QDomElement &&el) {
@@ -826,8 +831,9 @@ QXmppClient::State QXmppClient::state() const
 {
     if (d->stream->isConnected()) {
         return QXmppClient::ConnectedState;
-    } else if (d->stream->socket()->state() != QAbstractSocket::UnconnectedState &&
-               d->stream->socket()->state() != QAbstractSocket::ClosingState) {
+    }
+    auto state = d->stream->xmppSocket().internalSocket()->state();
+    if (state != QAbstractSocket::UnconnectedState && state != QAbstractSocket::ClosingState) {
         return QXmppClient::ConnectingState;
     } else {
         return QXmppClient::DisconnectedState;
@@ -879,13 +885,13 @@ void QXmppClient::setClientPresence(const QXmppPresence &presence)
 /// Returns the socket error if error() is QXmppClient::SocketError.
 QAbstractSocket::SocketError QXmppClient::socketError()
 {
-    return d->stream->socket()->error();
+    return d->stream->xmppSocket().internalSocket()->error();
 }
 
 /// Returns the human-readable description of the last socket error if error() is QXmppClient::SocketError.
 QString QXmppClient::socketErrorString() const
 {
-    return d->stream->socket()->errorString();
+    return d->stream->xmppSocket().internalSocket()->errorString();
 }
 
 /// Returns the XMPP stream error if QXmppClient::Error is QXmppClient::XmppStreamError.
@@ -931,6 +937,18 @@ bool QXmppClient::injectMessage(QXmppMessage &&message)
     return handled;
 }
 
+///
+/// Sets stream errors that are ignored if they occur.
+///
+/// \param errors stream errors to be ignored
+///
+/// \since QXmpp 1.9
+///
+void QXmppClient::setIgnoredStreamErrors(const QVector<QXmpp::StreamError> &errors)
+{
+    d->ignoredStreamErrors = errors;
+}
+
 ///
 /// Give extensions a chance to handle incoming stanzas.
 ///
@@ -950,9 +968,9 @@ void QXmppClient::_q_reconnect()
     }
 }
 
-void QXmppClient::_q_socketStateChanged(QAbstractSocket::SocketState socketState)
+void QXmppClient::onInternalSocketStateChanged()
 {
-    Q_UNUSED(socketState);
+    // FIXME: only emit signal if state actually changed
     Q_EMIT stateChanged(state());
 }
 
